"""
Lavalink connecter
~~~~~~~~~~~~~~~
:copyright: (c) 2021 HazemMeqdad
:license: MIT, see LICENSE for more details.
"""

__title__ = "lavaplayer"
__author__ = "HazemMeqdad"
__license__ = "MIT"
__version__ = "1.0.4a"

from .client import LavalinkClient
from .objects import *
from .api import Api
from .exceptions import *